from random import shuffle
from unittest import TestCase

from mercure.client.publisher.base import BasePublisherClient


class BasePublisherClientTestCase(TestCase):
    def setUp(self):
        self.client = BasePublisherClient(hub_url='http://localhost',
                                          channel_generation_salt='yammy')
    
    def test_get_channel_name(self):
        expected = (
            '73d6cd526857b9a0c9f7e3d9b7eed5db730b13a14e6f889081064019f5ccc17c'
        )

        tags = ['any', 1, 'of', 'objs']
        channel_name = self.client.get_channel_name(*tags)
        self.assertEqual(
            channel_name,
            expected,
        )

        shuffle(tags)
        # check that any order gives same result
        channel_name = self.client.get_channel_name(*tags)
        self.assertEqual(
            channel_name,
            expected,
        )



from asyncio import create_task

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientError

from mercure.client.publisher.base import (
    DEFAULT_EVENT_TYPE_MAX_SIZE,
    DEFAULT_EVENT_MESSAGE_MAX_SIZE,
    BasePublisherClient,
)


class MercurePublisherClientAsync(BasePublisherClient):
    def __init__(self,
                 hub_url,
                 event_type_max_size=DEFAULT_EVENT_TYPE_MAX_SIZE,
                 event_message_size=DEFAULT_EVENT_MESSAGE_MAX_SIZE,
                 channel_generation_salt='',
                 transport_session=None):
        super().__init__(hub_url, event_type_max_size, event_message_size,
                         channel_generation_salt)
        self.session = transport_session or ClientSession()

    async def _request(self, channel_name, payload):
        try:
            await self.session.post(
                url=self.hub_url,
                params={'id': channel_name},
                data=payload,
            )
        except ClientError:
            # We don't guarantee delivery, failing silently
            pass

    def _invoke_request(self, channel_name, payload):
        create_task(self._request(channel_name, payload))



from gevent import spawn

from mercure.client.publisher.common import MercurePublisherClient


class MercurePublisherClientGevent(MercurePublisherClient):
    def _invoke_request(self, channel_name, payload):
        spawn(super()._invoke_request, channel_name, payload)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mercure',
 'mercure.client',
 'mercure.client.publisher',
 'mercure.client.publisher.tests',
 'mercure.client.subscriber']

package_data = \
{'': ['*']}

extras_require = \
{'async': ['aiohttp>=3.4.4,<4.0.0'],
 'common': ['requests>=2.21,<3.0'],
 'gevent': ['requests>=2.21,<3.0', 'gevent>=1.4.0,<2.0.0']}

setup_kwargs = {
    'name': 'mercure',
    'version': '0.0.2.0',
    'description': 'The service to push events between parties. The messenger of gods.',
    'long_description': None,
    'author': 'BCD TripTech',
    'author_email': 'development@bcdtriptech.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
from hashlib import sha256
from json import dumps
from urllib.parse import urljoin


DEFAULT_EVENT_TYPE_MAX_SIZE = 30
DEFAULT_EVENT_MESSAGE_MAX_SIZE = 100
CHANNEL_NAME_MAX_SIZE = 200
# This is the exact number of extra characters in the chosen message structure
# after json-serialization
PAYLOAD_SERIALIZATION_OVERHEAD = 39


class BasePublisherClient(metaclass=ABCMeta):
    """
    Base Mercure client class.
    
    Sub-classes must implement BasePublisherClient._invoke_request.
    """
    def __init__(
        self, 
        hub_url,
        event_type_max_size=DEFAULT_EVENT_TYPE_MAX_SIZE,
        event_message_max_size=DEFAULT_EVENT_MESSAGE_MAX_SIZE,
        channel_generation_salt='',
    ):
        self.hub_url = urljoin(hub_url, 'pub')
        self.channel_generation_salt = channel_generation_salt

        # These restrictions are purely artificial: they are added to
        # somewhat enforce anonymization of payload and avoiding adding
        # any auth for the service.
        self.event_type_max_size = event_type_max_size
        self.event_message_max_size = event_message_max_size

    def get_channel_name(self, *tags):
        """
        Generate a channel name for specific set of 'tags'. 

        Tags are any number of objects that can be cast to str.

        TODO: DOS is possible if attacker subscribes to lots and lots of
        channels and eats nginx's memory. To prevent this we might want
        to forbid subscribers create channels and create channel in the hub
        somewhere around channel name generation step.
        """
        # Deterministic list of tags:
        ordered_tags = sorted([
            self.channel_generation_salt, 
            *map(str, tags),
        ])
        return sha256('|'.join(ordered_tags).encode()).hexdigest()

    def _validate_params_and_prepare_payload(self, channel_name, 
                                             event_type, event_message):
        if not isinstance(channel_name, str):
            raise ValueError('"channel_name" should be a string')
        if len(channel_name) > CHANNEL_NAME_MAX_SIZE:
            raise ValueError(f'"channel_name" size should not exceed '
                             f'{CHANNEL_NAME_MAX_SIZE}')

        if not isinstance(event_type, str):
            raise ValueError('"event_type" should be a string')
        event_type_len = len(event_type)
        if event_type_len > self.event_type_max_size:
            raise ValueError(f'"event_type" size should not exceed '
                             f'{self.event_type_max_size}')

        payload = {'event_type': event_type, 'event_message': event_message}

        try:
            payload = dumps(payload)
        except TypeError:
            raise TypeError('"event_message" should be JSON serializable')

        message_size = (
            len(payload) - event_type_len - PAYLOAD_SERIALIZATION_OVERHEAD
        )

        if message_size > self.event_message_max_size:
            raise ValueError(f'"event_message" size should not exceed '
                             f'{self.event_message_max_size} in'
                             f'JSON serialized form')

        return payload

    @abstractmethod
    def _invoke_request(self, channel_name, payload):
        pass

    def publish_message(self, channel_name, event_type, event_message):
        """
        Publish a message to a given channel name.

        It doesn't allow to send complicated and big messages on purpose:
        we want to avoid authentication on the hub, thus only depersonalized
        data should be transferred through it.
        Channel name could be an arbitrary string or generated one using
        MercureClient().get_channel_name().
        
        WARNING: It doesn't provide any guarantees on message delivery, thus
        will silently fail on any network errors.
        """
        payload = self._validate_params_and_prepare_payload(
                channel_name,
                event_type,
                event_message,
        )
        self._invoke_request(channel_name, payload)

    def publish_message_by_tags(self, tags, event_type, event_message):
        """
        Generate a channel name from tags and push a message into it.

        It doesn't allow to send complicated and big messages on purpose:
        we want to avoid authentication on the hub, thus only depersonalized
        data should be transferred through it.
        
        WARNING: It doesn't provide any guarantees on message delivery, thus
        will silently fail on any network errors.
        """
        channel_name = self.get_channel_name(*tags)
        self.publish_message(channel_name, event_type, event_message)
